/***************************************************************/
/*							       */
/*			     POLKAWIDGETMOTIF.C		       */
/*							       */
/***************************************************************/
/*       Copyright 1993 Georgia Institute of Technology 
                     -- John T. Stasko                         */

#include "polka_local.H"
#include "polka_version.H"
#include "config.h"

#include <sys/types.h>	// For micro_sleep() 
#include <sys/time.h>	// For micro_sleep() 
#include <sys/socket.h>	// For micro_sleep() 
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
/* On some systems include <sys/select.h> instead of the three above */
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#ifndef SELECT_DEFINED
extern "C" {
 extern int select(long unsigned int, void *, void *, void *, struct timeval *);
}
#endif

#include <X11/StringDefs.h>
#include <X11/Shell.h>

#include <Xm/Xm.h>
#include <Xm/Form.h>
#include <Xm/DrawingA.h>
#include <Xm/Label.h>
#include <Xm/RowColumn.h>
#include <Xm/PushB.h>
#include <Xm/CascadeB.h>
#include <Xm/ToggleB.h>
#include <Xm/ArrowB.h>
#include <Xm/ScrollBar.h>


/* Variables and functions for Control Panel */

const  int          SLIDERSIZE = 5;
       Display      *_display = NULL;
       XtAppContext  _context = NULL;
       int           _color_screen;
       int           _delay = 0;
static int           _started = 0;
static int           _paused;
static Widget        Runbut;             // needed by stepCB
static Widget        Speedbar;           // needed by PolkaReset
/*** COMMENT: some static variables and procedures are made global so
     that subclasses defined in another file can access them. ***/
Widget        _appshell;

static void controlPanel();
static void speedCB(Widget, XtPointer, XtPointer);
static void pauseCB(Widget, XtPointer, XtPointer);
static void stepCB(Widget, XtPointer, XtPointer);



/* Functions for View windows */

void debugCB(Widget, XtPointer, XtPointer);
void refreshCB(Widget, XtPointer, XtPointer);
void closeCB(Widget, XtPointer, XtPointer);
static void panlCB(Widget, XtPointer, XtPointer);
static void panrCB(Widget, XtPointer, XtPointer);
static void pandCB(Widget, XtPointer, XtPointer);
static void panuCB(Widget, XtPointer, XtPointer);
static void zoominCB(Widget, XtPointer, XtPointer);
static void zoomoutCB(Widget, XtPointer, XtPointer);


/* Functions that the driver application calling Polka must define */


APP_CALLBACK AppRunPauseCB = NULL;
APP_CALLBACK AppStepCB = NULL;
APP_CALLBACK AppDelayCB = NULL;
APP_CALLBACK AppCloseCB = NULL;


void
PolkaInitialize(Display *d, XtAppContext c, Widget a)
{
   _display = d;
   _context = c;
   _appshell = a;
}


void
PolkaSetCallbacks(APP_CALLBACK rp, APP_CALLBACK st, APP_CALLBACK del,
   APP_CALLBACK clo)
{
   AppRunPauseCB = rp;
   AppStepCB = st;
   AppDelayCB = del;
   AppCloseCB = clo;
}


void
PolkaReset()
{
   Arg args[1];
   XmString newLabel;

   _paused = 1;
   newLabel = XmStringCreate((char*)" Start ",XmSTRING_DEFAULT_CHARSET);
   XtSetArg(args[0], XmNlabelString, newLabel);
   XtSetValues(Runbut, args, 1);
   XmStringFree(newLabel);

   XmScrollBarSetValues(Speedbar, 0, SLIDERSIZE, 1, 10, True);
   _delay = 0;
}



void
PolkaInit()
{
   Visual *vis;
   int argc = 1;  // need to fake out XtDisplayInitialize
   char *argv[1];
   static const char *fallback_resources[] = {
           "*animwindow*width:    512",
           "*animwindow*height:    512",
           NULL  };
   /* may need to make that const instead of static */

   if (_started) return;

   _paused = 1;

   if (!_display) {
      argv[0] = (char*)"__Polka";
      _appshell = XtAppInitialize(&_context, "Polka", NULL, 0,
           &argc, argv, (char**)fallback_resources, NULL, 0);
      _display = XtDisplay(_appshell);
      }

   printf("\nPOLKA Version %4.02f\n\n", POLKAVERSION);
   vis = DefaultVisual(_display,DefaultScreen(_display));
   _color_screen = ((vis->c_class != GrayScale) && 
                       (vis->c_class != StaticGray));

   initGC();
   controlPanel();
   _started = 1;
}



void
controlPanel()
{
   Widget      top, rowcol, stepbut, fastlabel, slowlabel;
   XmString    xmstr;
   Arg	       args[20];
   int         n;

      n = 0;
      XtSetArg(args[n], XmNtitle, "Polka Control Panel"); n++;
   top = XtCreatePopupShell("ControlPanel", topLevelShellWidgetClass, 
                _appshell, args, n);

      n = 0;
      XtSetArg(args[n], XmNorientation, XmHORIZONTAL); n++;
      XtSetArg(args[n], XmNresizeWidth, True); n++;
      XtSetArg(args[n], XmNspacing, 5); n++;
      XtSetArg(args[n], XmNmarginWidth, 5); n++;
   rowcol = XtCreateManagedWidget("rowcol", xmRowColumnWidgetClass, 
                   top, args, n);

      n = 0;
      xmstr = XmStringCreate((char*)" Start ",XmSTRING_DEFAULT_CHARSET); 
      XtSetArg(args[n], XmNlabelString, xmstr); n++;
      XtSetArg(args[n], XmNrecomputeSize, True); n++;
   Runbut = XtCreateManagedWidget("runbut", xmPushButtonWidgetClass,
                      rowcol, args, n);
      XtAddCallback(Runbut, XmNactivateCallback, pauseCB, NULL);
      if (AppRunPauseCB)
         XtAddCallback(Runbut, XmNactivateCallback, *AppRunPauseCB, NULL);

      n = 0;
      xmstr = XmStringCreate((char*)"Step",XmSTRING_DEFAULT_CHARSET); 
      XtSetArg(args[n], XmNlabelString, xmstr); n++;
   stepbut = XtCreateManagedWidget("stepbut", xmPushButtonWidgetClass,
                      rowcol, args, n);
      XtAddCallback(stepbut, XmNactivateCallback, stepCB, NULL);
      if (AppStepCB) 
         XtAddCallback(stepbut, XmNactivateCallback, *AppStepCB, NULL);

      n = 0;
      xmstr = XmStringCreate((char*)"Slow",XmSTRING_DEFAULT_CHARSET); 
      XtSetArg(args[n], XmNlabelString, xmstr); n++;
   slowlabel = XtCreateManagedWidget("slowlabel", xmLabelWidgetClass, 
                        rowcol, args, n);
   (void)slowlabel;

      n = 0;
      XtSetArg(args[n], XmNorientation, XmHORIZONTAL); n++;
      XtSetArg(args[n], XmNprocessingDirection, XmMAX_ON_LEFT); n++;
      XtSetArg(args[n], XmNsliderSize, SLIDERSIZE); n++;
   Speedbar = XtCreateManagedWidget("speedbar", xmScrollBarWidgetClass, 
                     rowcol, args, n);
      XtAddCallback(Speedbar, XmNvalueChangedCallback, speedCB, NULL);
      if (AppDelayCB) {
         XtAddCallback(Speedbar, XmNvalueChangedCallback, *AppDelayCB, NULL);
         XtAddCallback(Speedbar, XmNdragCallback, *AppDelayCB, NULL);
         }


      n = 0;
      xmstr = XmStringCreate((char*)"Fast",XmSTRING_DEFAULT_CHARSET); 
      XtSetArg(args[n], XmNlabelString, xmstr); n++;
   fastlabel = XtCreateManagedWidget("fastlabel", xmLabelWidgetClass, 
                        rowcol, args, n);

   (void)fastlabel;
   XtPopup(top, XtGrabNone);
}



/**************************************************************/
/* CheckXEvents -- Handle all pending X events.		      */
/*							      */
/* RETURNS:  None.					      */
/**************************************************************/

void
CheckXEvents()
{
   XEvent xev;

   while (XtAppPending(_context)) {
      XtAppNextEvent(_context, &xev);
      XtDispatchEvent(&xev);
      }
}



/**************************************************************/
/* BatchMode - FLush out X graphics.                          */
/*							      */
/* RETURNS:  None.					      */
/**************************************************************/
void
BatchMode(int fg)
{
   static int batch_level = 0;

   if (fg)
      ++batch_level;
   else if (batch_level > 0) {
      --batch_level;
      if (batch_level == 0)
	 XFlush(_display);
      }
}




/***************************************************************/
/* micro_sleep - used to busy wait and slow down the animation.*/
/*	   This is the most machine specific code in the       */
/*         package and may not compile/work on your machine.   */
/*							       */
/* RETURNS:  None.					       */
/***************************************************************/

void 
micro_sleep(int usec)
{
   struct timeval t;

   t.tv_sec = usec >> 20;
   t.tv_usec = usec & 0xfffff;
   if (t.tv_usec >= 1000000) {
      ++t.tv_sec;
      t.tv_usec -= 1000000;
      }

   select(0, 0, 0, 0, &t);	/* Block until time expires */

//  Alternative (ultra low-tech way).  If above doesn't compile,
//  comment it out and uncomment this code.
//  for (int i=0; i<usec; i++)
//     continue;
//
}



/***************************************************************/
/* speedCB -- Adjust the delay setting according               */
/*	       to the scrollbar position.		       */
/*							       */
/* RETURNS:  None.					       */
/***************************************************************/

void
speedCB(Widget, XtPointer, XtPointer sl)
{
   XmScrollBarCallbackStruct *slider = (XmScrollBarCallbackStruct *)sl;
   _delay = 2000 * slider->value;
}



/**************************************************************/
/* pauseCB -- Pause/unpause animation and change	      */
/*		     button label to reflect current state.   */
/*							      */
/* RETURNS:  None.					      */
/**************************************************************/
void
pauseCB(Widget w, XtPointer, XtPointer)
{
   Arg args[1];
   XmString newLabel;

   // Initially paused as "run animation" 
   if (_paused) {
      _paused = 0;
      }
   else
      _paused = 1;
   newLabel = XmStringCreate((char*)(_paused ? "Run" : "Pause"),
                  XmSTRING_DEFAULT_CHARSET);
   XtSetArg(args[0], XmNlabelString, newLabel);
   XtSetValues(w, args, 1);
   XmStringFree(newLabel);
}



/**************************************************************/
/* stepCB -- Step one animation frame			      */
/*							      */
/* RETURNS:  None.					      */
/**************************************************************/

void
stepCB(Widget, XtPointer, XtPointer)
{
   Arg args[1];
   XmString newLabel;

   if (!_paused) {
      _paused = 1;
      newLabel = XmStringCreate((char*)"Run",XmSTRING_DEFAULT_CHARSET);
      XtSetArg(args[0], XmNlabelString, newLabel);
      XtSetValues(Runbut, args, 1);
      XmStringFree(newLabel);
      }
}










/*---------------------------------------------------------------------------*/
/*                              POLKA Widgets                                */
/*---------------------------------------------------------------------------*/
/*                                                                           */
/*                             toplevel                                      */
/*                                |                                          */
/*                               form                                        */
/*       /---------------/--------+---\-------\-----\-----\------\-----\     */
/*     title           rcleft         sbar   pause mode debug refresh quit   */
/*               /----/---\---\---\--\                                       */
/*            left right up down in out        			             */
/*                                                                           */
/*                                                                           */
/*	modeMgrShell			debugMgrShell		      */
/*	      |				      |			      */
/*	modeMenuMgr			debugMenuMgr		      */
/*	      |				      |			      */
/*	 modeShell			 debugShell		      */
/*	      |				      |			      */
/*	  Refresh			    Debug		      */
/*    /---------|-------\		    /-+-\		      */
/* by frame  by scene  none		  *Off  On		      */
/*---------------------------------------------------------------------------*/

/**************************************************************/
/* setupWindow -- Create, init (hardcoded), place, &          */
/*		  realize all widgets for animation windows.  */
/*							      */
/* RETURNS:  None.					      */
/**************************************************************/
void
View::setupWindow(const char *title)
{
   Widget      form, botform, botrightrowcol, debugbut,
                refreshbut, closebut,
                botleftrowcol, larrowbut, rarrowbut, uarrowbut, darrowbut, 
                zoominbut, zoomoutbut;
   XmString    xmstr;
   Arg	       args[20];
   int         n;
   APP_CALLBACK cb;

   if (debug)
      printf("DEBUG: View::setupWindow\n");

      n = 0;
      XtSetArg(args[n], XmNtitle, title); n++;
   topshell = XtCreatePopupShell("AnimWindow", topLevelShellWidgetClass, 
                _appshell, args, n);

   form = XtCreateManagedWidget("topform", xmFormWidgetClass, topshell,   
                   NULL, 0);

      n = 0;
      XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
   botform = XtCreateManagedWidget("botform", xmFormWidgetClass, form,
                                     args, n);

      n = 0;
      XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNorientation, XmHORIZONTAL); n++;
   botleftrowcol = XtCreateManagedWidget("botleftrowcol", 
                         xmRowColumnWidgetClass, botform, args, n);

      n = 0;
      XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNbottomAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNorientation, XmHORIZONTAL); n++;
   botrightrowcol = XtCreateManagedWidget("botrightrowcol", 
                       xmRowColumnWidgetClass, botform, args, n);

      n = 0;
      xmstr = XmStringCreate((char*)"Debug",XmSTRING_DEFAULT_CHARSET); 
      XtSetArg(args[n], XmNlabelString, xmstr); n++;
      XtSetArg(args[n], XmNindicatorOn, False); n++;
   debugbut = XtCreateManagedWidget("debugbut", xmToggleButtonWidgetClass, 
                             botrightrowcol, args, n);
      XtAddCallback(debugbut, XmNvalueChangedCallback, debugCB, (XtPointer)this);

      n = 0;
      xmstr = XmStringCreate((char*)"Refresh",XmSTRING_DEFAULT_CHARSET); 
      XtSetArg(args[n], XmNlabelString, xmstr); n++;
   refreshbut = XtCreateManagedWidget("modebut", xmPushButtonWidgetClass, 
                             botrightrowcol, args, n);
      XtAddCallback(refreshbut, XmNactivateCallback, refreshCB, (XtPointer)this);

      n = 0;
      xmstr = XmStringCreate((char*)"Close",XmSTRING_DEFAULT_CHARSET); 
      XtSetArg(args[n], XmNlabelString, xmstr); n++;
   closebut = XtCreateManagedWidget("closebut", xmPushButtonWidgetClass,
                            botrightrowcol, args, n);
      XtAddCallback(closebut, XmNactivateCallback, closeCB, 
                       (XtPointer)topshell);
      if (AppCloseCB)
         XtAddCallback(closebut, XmNactivateCallback, *AppCloseCB, 
                       (XtPointer)this);

      n = 0;
      XtSetArg(args[n], XmNarrowDirection, XmARROW_LEFT); n++;
   larrowbut = XtCreateManagedWidget("larrowbut", xmArrowButtonWidgetClass, 
				     botleftrowcol, args, n);
      if (AppPanLeftCB) cb = AppPanLeftCB;
      else cb = &panlCB;
      XtAddCallback(larrowbut, XmNactivateCallback, *cb, (XtPointer)this);

      n = 0;
      XtSetArg(args[n], XmNarrowDirection, XmARROW_RIGHT); n++;
   rarrowbut = XtCreateManagedWidget("rarrowbut", xmArrowButtonWidgetClass, 
                                 botleftrowcol, args, n);
      if (AppPanRightCB) cb = AppPanRightCB;
      else cb = &panrCB;
      XtAddCallback(rarrowbut, XmNactivateCallback, *cb, (XtPointer)this);

      n = 0;
      XtSetArg(args[n], XmNarrowDirection, XmARROW_DOWN); n++;
   darrowbut = XtCreateManagedWidget("darrowbut", xmArrowButtonWidgetClass, 
                                 botleftrowcol, args, n);
      if (AppPanDownCB) cb = AppPanDownCB;
      else cb = &pandCB;
      XtAddCallback(darrowbut, XmNactivateCallback, *cb, (XtPointer)this);

      n = 0;
      XtSetArg(args[n], XmNarrowDirection, XmARROW_UP); n++;
   uarrowbut = XtCreateManagedWidget("uarrowbut", xmArrowButtonWidgetClass, 
                                 botleftrowcol, args, n);
      if (AppPanUpCB) cb = AppPanUpCB;
      else cb = &panuCB;
      XtAddCallback(uarrowbut, XmNactivateCallback, *cb, (XtPointer)this);

      n = 0;
      xmstr = XmStringCreate((char*)"In",XmSTRING_DEFAULT_CHARSET); 
      XtSetArg(args[n], XmNlabelString, xmstr); n++;
   zoominbut = XtCreateManagedWidget("zoominbut", xmPushButtonWidgetClass, 
                                 botleftrowcol, args, n);
      if (AppZoomInCB) cb = AppZoomInCB;
      else cb = &zoominCB;
      XtAddCallback(zoominbut, XmNactivateCallback, *cb, (XtPointer)this);

      n = 0;
      xmstr = XmStringCreate((char*)"Out",XmSTRING_DEFAULT_CHARSET); 
      XtSetArg(args[n], XmNlabelString, xmstr); n++;
   zoomoutbut = XtCreateManagedWidget("zoomoutbut", xmPushButtonWidgetClass, 
                                 botleftrowcol, args, n);
      if (AppZoomOutCB) cb = AppZoomOutCB;
      else cb = &zoomoutCB;
      XtAddCallback(zoomoutbut, XmNactivateCallback, *cb, (XtPointer)this);

      n = 0;
      XtSetArg(args[n], XtNborderWidth, 1); n++;
      XtSetArg(args[n], XmNresizePolicy, XmRESIZE_ANY); n++;
      XtSetArg(args[n], XmNtopAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNbottomAttachment, XmATTACH_WIDGET); n++;
      XtSetArg(args[n], XmNbottomWidget, botform); n++;
      XtSetArg(args[n], XmNleftAttachment, XmATTACH_FORM); n++;
      XtSetArg(args[n], XmNrightAttachment, XmATTACH_FORM); n++;
      if ((pixX > 0) && (pixY > 0)) {   // set by the user via View::Create
         XtSetArg(args[n], XmNwidth, pixX); n++;
         XtSetArg(args[n], XmNheight, pixY); n++;
      }
   easel = XtCreateManagedWidget("animwindow", xmDrawingAreaWidgetClass, form,
				 args, n);
      XtAddCallback(easel, XmNexposeCallback, resize, (XtPointer)this);
      XtAddCallback(easel, XmNresizeCallback, resize, (XtPointer)this);
      XtAddEventHandler(easel, ButtonPressMask, FALSE, SelectPoint, 
                     (XtPointer)this);

   XtPopup(topshell, XtGrabNone);
}


/**************************************************************/
/* SetPanCallbacks -- set user supplied callback routines     */
/*           for panning and zooming.                         */
/*							      */
/* RETURNS:  None.					      */
/**************************************************************/
void
View::SetPanCallbacks(APP_CALLBACK lmove, APP_CALLBACK rmove,
		      APP_CALLBACK umove, APP_CALLBACK dmove,
		      APP_CALLBACK zoomin, APP_CALLBACK zoomout)
{
    AppPanLeftCB = lmove;
    AppPanRightCB = rmove;
    AppPanUpCB = umove;
    AppPanDownCB = dmove;
    AppZoomInCB = zoomin;
    AppZoomOutCB = zoomout;
}


/**************************************************************/
/* refreshCB -- (callback function) Redraw all images         */
/*           onscreen.                                        */
/*							      */
/* RETURNS:  None.					      */
/**************************************************************/
void
refreshCB(Widget, XtPointer thisone, XtPointer)
{
   ((View *)thisone)->Refresh();
}



/**************************************************************/
/* closeCB -- close this view				      */
/*							      */
/* RETURNS:  None.					      */
/**************************************************************/
void
closeCB(Widget, XtPointer t, XtPointer)
{
   Widget topshell = (Widget) t;
   XtPopdown(topshell);
}



/**************************************************************/
/* resize -- Resize widgets.				      */
/*							      */
/* RETURNS:  None.					      */
/**************************************************************/
void
resize(Widget w, XtPointer thisone, XtPointer)
{
   View *t = (View *)thisone;
   if (t->window) {	/* Avoid resize when widget initially displayed */
		/* -- *before* offscreen pixmap has been created.	*/
      t->prepareAnim(XtWindow(w));
      t->fixPad();
      t->Refresh();
      }
}



/**************************************************************/
/* changeDebug_callback -- Change debug mode.   	      */
/*			   				      */
/*							      */
/* RETURNS:  None.					      */
/**************************************************************/
void
debugCB(Widget, XtPointer client_data, XtPointer)
{
   View *v;
   int d;

   v = (View *)client_data;
   d = v->GetDebug();
   v->SetDebug( d ? 0 : 1);
}






/**************************************************************/
/* panning and zooming callbacks.		    	      */
/*							      */
/* RETURNS:  None.					      */
/**************************************************************/

static void
panlCB(Widget, XtPointer client_data, XtPointer)
{
   View *v = (View *)client_data;
   double lx,by,rx,ty,change,panfactor = 0.2;

   v->GetCoord(lx,by,rx,ty);
   change = (rx-lx) * panfactor;
   v->SetCoord(lx-change, by, rx-change, ty);
}


static void
panrCB(Widget, XtPointer client_data, XtPointer)
{
   View *v = (View *)client_data;
   double lx,by,rx,ty,change,panfactor = 0.2;

   v->GetCoord(lx,by,rx,ty);
   change = (rx-lx) * panfactor;
   v->SetCoord(lx+change, by, rx+change, ty);
}


static void
pandCB(Widget, XtPointer client_data, XtPointer)
{
   View *v = (View *)client_data;
   double lx,by,rx,ty,change,panfactor = 0.2;

   v->GetCoord(lx,by,rx,ty);
   change = (ty-by) * panfactor;
   v->SetCoord(lx, by-change, rx, ty-change);
}


static void
panuCB(Widget, XtPointer client_data, XtPointer)
{
   View *v = (View *)client_data;
   double lx,by,rx,ty,change,panfactor = 0.2;

   v->GetCoord(lx,by,rx,ty);
   change = (ty-by) * panfactor;
   v->SetCoord(lx, by+change, rx, ty+change);
}



static void
zoominCB(Widget, XtPointer client_data, XtPointer)
{
   View *v = (View *)client_data;
   double lx,by,rx,ty,changex,changey,zoomfactor = 0.8;

   v->GetCoord(lx,by,rx,ty);
   changey = ((ty-by) * (1.0-zoomfactor)) / 2.0;
   changex = ((rx-lx) * (1.0-zoomfactor)) / 2.0;
   v->SetCoord(lx+changex, by+changey, rx-changex, ty-changey);
}




static void
zoomoutCB(Widget, XtPointer client_data, XtPointer)
{
   View *v = (View *)client_data;
   double lx,by,rx,ty,changex,changey,zoomfactor = 0.8;

   v->GetCoord(lx,by,rx,ty);
   changey = ((ty-by) * (1.0/zoomfactor -1.0)) / 2.0;
   changex = ((rx-lx) * (1.0/zoomfactor -1.0)) / 2.0;
   v->SetCoord(lx-changex, by-changey, rx+changex, ty+changey);
}





