%{
#include <stdio.h>
#include <string.h>
#include "sensor.h"

/* extern char yytext[]; */
extern int yylineno;

extern ValidSensor();
extern ValidAttribute();

SensorP		sensors = NULL;
int		numSensors = 0;

int             errors = 0;
%}
%union{
    SensorP        NTSensorType;
    AttributeP     NTAttributeType;
    AttributeType  NTSimpleType;
    LocationP      NTLocationType;
    PolicyP	   NTPolicyType;
    NumberListP    NTNumberListType;
    char*	   NTStringType;
    int		   NTint;
    char	   nothing;
}


%token EVENTSENSOR
%token STATESENSOR
%token SENSOR
%token ATTRIBUTE
%token LOCATION
%token POLICY

%token DEFAULT
%token USER_SPECIFY
%token VERBATIM
%token BEFORE_PROC_CALL
%token AFTER_PROC_CALL
%token BEGINNING_PROC
%token END_PROC
%token BEFORE_SEND 
%token AFTER_SEND
%token BEFORE_RECEIVE
%token AFTER_RECEIVE
%token BEFORE_SH_ACCESS
%token AFTER_SH_ACCESS
%token AT_LINES
%token IN_FILE

%token EXPRESSION
%token NUMBER
%token ID

%token CHAR
%token SHORT
%token INT
%token LONG
%token UCHAR
%token USHORT
%token UINT
%token ULONG
%token FLOAT
%token DOUBLE
%token STRING
%token CHARSTAR

%token COLON
%token COMMA
%token SEMICOLON
%token EQUAL
%token LPARAN
%token RPARAN
%token LBRACKET
%token RBRACKET


%type <NTSensorType>		_sensorspec
%type <NTSensorType >		_sensors
%type <NTSensorType>   		_statedef
%type <NTSensorType>		_eventdef

%type <NTAttributeType>     	_attributes
%type <NTAttributeType>     	_attribute_list
%type <NTAttributeType>		_attribute
%type <NTint>			_attrtype
%type <NTSimpleType>		_simple_type _string_type
%type <NTint>			_array_size
%type <NTStringType>		_attrexpr
%type <NTStringType>		_attribute_expression
%type <NTStringType>		_attribute_expr_part

%type <NTLocationType>		_locations
%type <NTLocationType>		_location_values
%type <NTLocationType>		_location_value
%type <NTNumberListType>	_lines
%type <NTNumberListType>	_line_number

%type <NTPolicyType>		_policies
%type <NTPolicyType>		_policy_values
%type <NTPolicyType>		_policy_value

%type <NTStringType>		_name


%%


_sensorspec :
        _sensors
	{
	    SensorP  p;
	    
	    p = sensors = $1;
	    while (p) {numSensors++; p = p->next;}
	    $$ = $1;
	}
        ;


_sensors :
        _eventdef
        {
	    $$ = $1;
        }
        |
        _statedef
	{
	    $$ = $1;
	}
        | 
        _eventdef _sensors
        {
	    if (!ValidSensor($1, $2)) errors++;
	    $2->prev = $1;
	    $1->next = $2;

	    $$ = $1;
        }
        |
        _statedef _sensors
	{
	    if (!ValidSensor($1, $2)) errors++;
	    $2->prev = $1;
	    $1->next = $2;

	    $$ = $1;
        }
        ;


_eventdef:
	EVENTSENSOR _name LPARAN _attributes  _locations  _policies RPARAN
  	_OPTIONAL_SEMI
        {
	    SensorP thisSensor;

	    SensorInit(thisSensor, $2, $4, $5, $6);
	    thisSensor->sensorType = EVENT_SENSOR;
	    $$ = thisSensor;
        }
	;


_eventdef:
	SENSOR _name LPARAN _attributes  _locations  _policies RPARAN
  	_OPTIONAL_SEMI
        {
	    SensorP thisSensor;

	    SensorInit(thisSensor, $2, $4, $5, $6);
	    thisSensor->sensorType = EVENT_SENSOR;
	    $$ = thisSensor;
        }
	;


_statedef:
        STATESENSOR _name LPARAN _attributes  _locations  _policies RPARAN
  	_OPTIONAL_SEMI
        {
	    SensorP thisSensor;

	    SensorInit(thisSensor, $2, $4, $5, $6);
	    thisSensor->sensorType = STATE_SENSOR;
	    $$ = thisSensor;
        }
        ;


_attributes:
	{
	    $$ = NULL;
	}
	|
	ATTRIBUTE  LPARAN  _attribute_list  RPARAN _OPTIONAL_SEMI
	{
	    $$ = $3;
	}
	;

_attribute_list:
	_attribute
        {
	    $$ = $1;
        }
        |
        _attribute _attribute_list
        {
	    if (!ValidAttribute($1, $2)) errors++;
	    $2->prev = $1;
	    $1->next = $2;

	    $$ = $1;
        }
	;


_attribute:
	_attrtype _name _attribute_expr_part
	{
	    short      type, array;
	    AttributeP thisAttr;

	    type = $1 % MAX_BASIC_TYPE;
	    array = $1 / MAX_BASIC_TYPE + 1;

	    AttributeInit(thisAttr, type, array, $2, $3);
	    $$ = thisAttr;
	} 
	;

_attribute_expr_part:
	SEMICOLON
	{
	    StringInit($$, "DEFAULT");
	}
	|
	COLON _attrexpr SEMICOLON
	{
	    $$ = $2;
	}
	|
	EQUAL _attrexpr SEMICOLON
	{
	    $$ = $2;
	}
	;

_attrtype :
	  _simple_type
	  {
              $$ = (int) $1;
          }
          |
          _simple_type LBRACKET _array_size RBRACKET
          {
	      int type;

	      type = (int) $1 + $3 * MAX_BASIC_TYPE;
	      $$ = type;
          }
	  |
	  _string_type
	  {
	      $$ = (int) $1;
	  }
          ;

_attrexpr :
	USER_SPECIFY
        {
	    StringInit($$, "USER_SPECIFY");
        }
	|
	DEFAULT
        {
	    StringInit($$, "DEFAULT");
        }
        |
	_attribute_expression
        {
	    $$ = $1;
        } 
        ;


_attribute_expression:
	EXPRESSION
        {
            StringInit($$, yytext);
        }
        ;


_array_size:
	NUMBER
	{
	    $$ = atoi(yytext);
	}
	;


_simple_type:
	CHAR
	{
	    $$ = CHAR_TYPE;
	}
	|
	SHORT
	{
	    $$ = SHORT_TYPE;
	}
	|
	INT
	{
	    $$ = INT_TYPE;
	}
	|
	LONG
	{
	    $$ = LONG_TYPE;
	}
	|
	UCHAR
	{
	    $$ = UCHAR_TYPE;
	}
	|
	USHORT
	{
	    $$ = USHORT_TYPE;
	}
	|
	UINT
	{
	    $$ = UINT_TYPE;
	}
	|
	ULONG
	{
	    $$ = ULONG_TYPE;
	}
	|
	FLOAT
	{
	    $$ = FLOAT_TYPE;
	}
	|
	DOUBLE
	{
	    $$ = DOUBLE_TYPE;
	}
        ;


_string_type:
	STRING
	{
	    $$ = STRING_TYPE;
	}
	|
	CHARSTAR
	{
	    $$ = STRING_TYPE;
	}
	;


_locations:
	{
	    $$ = NULL;
	}
	|
	LOCATION  _OPTIONAL_COLON  _location_values
	{
	    $$ = $3;
	}
	;

_location_values:
	_location_value
	{
	    $$ = $1;
	}
	|
	_location_value   _location_values
        {
	    $2->prev = $1;
	    $1->next = $2;

	    $$ = $1;
        }
	;

_location_value:
	DEFAULT  SEMICOLON
	{
	    LocationInit($$, DEFAULT_LOCATION, NULL, NULL);
	}
	|
	USER_SPECIFY  SEMICOLON
	{
	    LocationInit($$, USER_SPECIFY_LOCATION, NULL, NULL);
	}
	|
	BEFORE_PROC_CALL _name   SEMICOLON
	{
	    LocationInit($$, BEFORE_PROC_CALL_LOCATION, $2, NULL);
	}
	|
	AFTER_PROC_CALL _name   SEMICOLON
	{
	    LocationInit($$, AFTER_PROC_CALL_LOCATION, $2, NULL);
	}
	|
	BEGINNING_PROC  _name   SEMICOLON
	{
	    LocationInit($$, BEGINNING_PROC_LOCATION, $2, NULL);
	}
	|
	END_PROC  _name   SEMICOLON
	{
	    LocationInit($$, END_PROC_LOCATION, $2, NULL);
	}
	|
	BEFORE_SEND _name   SEMICOLON
	{
	    LocationInit($$, BEFORE_SEND_LOCATION, $2, NULL);
	}
	|
	AFTER_SEND _name   SEMICOLON
	{
	    LocationInit($$, AFTER_SEND_LOCATION, $2, NULL);
	}
	|
	BEFORE_RECEIVE _name   SEMICOLON
	{
	    LocationInit($$, BEFORE_RECEIVE_LOCATION, $2, NULL);
	}
	|
	AFTER_RECEIVE _name   SEMICOLON
	{
	    LocationInit($$, AFTER_RECEIVE_LOCATION, $2, NULL);
	}
	|
	BEFORE_SH_ACCESS _name   SEMICOLON
	{
	    LocationInit($$, BEFORE_SH_ACCESS_LOCATION, $2, NULL);
	}
	|
	AFTER_SH_ACCESS _name   SEMICOLON
	{
	    LocationInit($$, AFTER_SH_ACCESS_LOCATION, $2, NULL);
	}
	|
	AT_LINES  _lines  IN_FILE  _name   SEMICOLON
	{
	    LocationInit($$, AT_LINES_LOCATION, $4, $2);
	}
	;


_lines:
	_line_number
	{
	    $$ = $1;
	}
	|
	_line_number  COMMA  _lines
        {
	    $3->prev = $1;
	    $1->next = $3;

	    $$ = $1;
        }
	;


_line_number:
	NUMBER
	{
	    NumberListInit($$, atoi(yytext));
	}
	;


_policies:
	{
	    $$ = NULL;
	}
	|
	POLICY _OPTIONAL_COLON _policy_values
	{
	    $$ = $3;
	}
	;

_policy_values:
	_policy_value
	{
	    $$ = $1;
	}
	|
	_policy_value  _policy_values
        {
	    $2->prev = $1;
	    $1->next = $2;

	    $$ = $1;
        }
	;

_policy_value:
	DEFAULT  SEMICOLON
	{
	    PolicyInit($$, DEFAULT_POLICY);
	}
	;


_name:
	ID
        {
            StringInit($$, yytext);
        }
	;

_OPTIONAL_SEMI:
	|
	SEMICOLON
	;

_OPTIONAL_COLON:
	|
	COLON
	;

%%
#include "scanner.c"

#define ShowArgs \
  fprintf(stderr, "Usage %s -o <output file>\n", argv[0]); \
  fprintf(stderr, "         -event_format <evemt format file>\n"); \
  fprintf(stderr, "         -sensor_code <sensor code file>\n"); \
  fprintf(stderr, "         -hfile <sensors .h file>\n"); \
  fprintf(stderr, "         <input file>\n");


extern void OutSensorSpecification();

int  startSensorNumber = 128;

main(argc,argv)
int argc;
char *argv[]; 
{ 
    char   *outFileName = NULL, *inFileName = NULL;
    char   *eventFormatFileName = NULL;
    char   *sensorCodeFileName = NULL;
    char   *sensorHFileName = NULL;
    char   *sensorExtHFileName = NULL;
    char   *tempFileName;
    FILE   *outFile, *inFile, *eventFormatFile;
    FILE   *sensorCodeFile, *sensorHFile, *sensorExtHFile;
    short  i, l;
    char   m4Command[512];
    int    commandP = 0;

    inFileName = outFileName = eventFormatFileName = NULL;
    sprintf(m4Command, "m4 "); commandP = 3;
    
    /*** command line arguments processing ***/
    for (i = 1; i < argc; i++) {
	if (!strncmp(argv[i], "-o", 2)) {
	    /*** output file name ***/
	    if (i == argc-1) {
		ShowArgs;
		exit(1);
	    } else outFileName = argv[++i];
	    
	} else if (!strncmp(argv[i], "-event_format", 4)) {
	    /*** event format file name ***/
	    if (i == argc-1) {
		ShowArgs;
		exit(1);
	    } else  eventFormatFileName = argv[++i];

	} else if (!strncmp(argv[i], "-sensor_code", 4)) {
	    /*** sensor code file ***/
	    if (i == argc-1) {
		ShowArgs;
		exit(1);
	    } else  sensorCodeFileName = argv[++i];
	    
	} else if (!strncmp(argv[i], "-hfile", 4)) {
	    /*** sensors .h file ***/
	    if (i == argc-1) {
		ShowArgs;
		exit(1);
	    } else  sensorHFileName = argv[++i];

	    /*** options for m4 ***/
	} else if (!strcmp(argv[i], "-s") || !strcmp(argv[i], "-e")
		   || !strcmp(argv[i], "-se")) {
	    sprintf(&m4Command[commandP], "%s ", argv[i]);
	    commandP += strlen(argv[i])+1;
	} else if (!strncmp(argv[i], "-B", 2) || !strncmp(argv[i], "-H", 2)
		   || !strncmp(argv[i], "-S", 2) || !strncmp(argv[i], "-T", 2)
		   || !strncmp(argv[i], "-D", 2) || !strncmp(argv[i], "-U",2)) {
	    sprintf(&m4Command[commandP], "%s ", argv[i]);
	    commandP += strlen(argv[i])+1;
	    
	} else {
	    /*** input file name ***/
	    if (inFileName) {
		ShowArgs;
		exit(1);
	    } else {
		l = strlen(argv[i]);
		inFileName = (char*) malloc(l+3);
		strcpy(inFileName, argv[i]);
		if (!((l >= 2) && (inFileName[l-1] == 's')
		      && (inFileName[l-2] == '.'))) {
		    strcat(inFileName, ".s");
		    l += 2;
		}
	    }
	}
    }

    /*** invoke m4 and output to file with name tempFileName ***/
    if (inFileName) {
	tempFileName = (char*) malloc(strlen(inFileName)+16);
	sprintf(tempFileName, ".temp_input.%s", inFileName);
	sprintf(&m4Command[commandP], "%s > %s", inFileName, tempFileName);
    } else {
	tempFileName = ".temp_input_file";
	sprintf(&m4Command[commandP], "> %s", tempFileName);
    }

    system(m4Command);
    
    /*** open input file ***/
    if (!freopen(tempFileName,"r",stdin)) {
	fprintf(stderr, "Macro expassion error detected\n");
	exit(1);
    }
    
    /*** open  output file ***/
    if (!outFileName) {
	if (!inFileName) outFile = stdout;
	else {
	    outFileName = (char*) malloc(l+2);
	    strcpy(outFileName, inFileName);
	    outFileName[l] = 'f';
	    outFileName[l+1] = '\0';
	}
    }

    if (outFileName) {
	outFile = fopen(outFileName, "w+");
	if (!outFile) {
	    fprintf(stderr, "Cannot open %s\n", outFileName);
	    exit(1);
	}
    }
    
    /*** open event format file ***/
    if (!eventFormatFileName) {
	if (!inFileName) eventFormatFileName = "-";
	else {
	    eventFormatFileName = (char*) malloc(l+14);
	    strcpy(eventFormatFileName, inFileName);
	    eventFormatFileName[l-1] = '\0';
	    strcat(eventFormatFileName, "eventformat.c");
	}
    }
    
    /*** open sensor code file ***/
    if (!sensorCodeFileName) {
	if (!inFileName) sensorCodeFile = stdout;
	else {
	    sensorCodeFileName = (char*) malloc(l+13);
	    strcpy(sensorCodeFileName, inFileName);
	    sensorCodeFileName[l-1] = '\0';
	    strcat(sensorCodeFileName, "sensorcode.c");
	}
    }

    if (sensorCodeFileName) {
	sensorCodeFile = fopen(sensorCodeFileName, "w+");
	if (!sensorCodeFile) {
	    fprintf(stderr, "Cannot open %s\n", sensorCodeFileName);
	    exit(1);
	}
    }
    
    /*** open sensor .h file ***/
    if (!sensorHFileName) {
	if (!inFileName) sensorHFile = stdout;
	else {
	    sensorHFileName = (char*) malloc(l+12);
	    strcpy(sensorHFileName, inFileName);
	    sensorHFileName[l-1] = '\0';
	    strcat(sensorHFileName, "sensors.h");
	}
    }

    if (sensorHFileName) {
	sensorHFile = fopen(sensorHFileName, "w+");
	if (!sensorHFile) {
	    fprintf(stderr, "Cannot open %s\n", sensorHFileName);
	    exit(1);
	}
    }

    /*** open sensor extern H file ***/
    if (!inFileName) sensorExtHFile = stdout;
    else {
	sensorExtHFileName = (char*) malloc(l+16);
	strcpy(sensorExtHFileName, inFileName);
	sensorExtHFileName[l-1] = '\0';
	strcat(sensorExtHFileName, "sensorextern.h");
	sensorExtHFile = fopen(sensorExtHFileName, "w+");
	if (!sensorExtHFile) {
	    fprintf(stderr, "Cannot open %s\n", sensorExtHFileName);
	    exit(1);
	}
    }
    
    if (yyparse()) {
	fprintf(stderr, "\nSyntax error detected.\n");
	exit(1);
    } else if (errors) {
	fprintf(stderr, "\n%d error(s) detected.\n", errors);
	exit(1);
    }

    OutSensorSpecification(numSensors, sensors, outFile);
    fclose(outFile);

    GenerateEventFormat(numSensors, sensors, startSensorNumber,
			eventFormatFileName, sensorHFileName);

    GenerateSensorHFile(numSensors, sensors, startSensorNumber,
			sensorHFile, sensorHFileName);
    
    GenerateSensorCodeFile(numSensors, sensors, startSensorNumber,
		       sensorCodeFile, sensorHFileName);

    GenerateSensorExtHFile(numSensors, sensors, startSensorNumber,
			   sensorExtHFile);
    
    fclose(sensorCodeFile);
    fclose(sensorHFile);
    fclose(sensorExtHFile);

    exit(0);
}


yyerror()
{
  fprintf(stderr, "### syntax error at line: %d, near string: \"%s\" ###\n",
	  yylineno, yytext);
}
