%{
#include "config.h"
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include "sensor.h"

#ifdef YYTEXT_POINTER
extern char* yytext;
#else
extern char yytext[];
#endif
extern int yyerror();
extern int yy_line_num;
extern char *name;
extern char *expression;
extern int number;
extern int yylex();

extern ValidSensor();
extern ValidAttribute();

extern LocationP LocationInit(/* loc_type, loc_name, loc_lines */);
extern AttributeP AttributeInit(/* attr_type, attr_array_size, attr_name,
				   attr_expr */);
extern SensorP SensorInit(/* sensor_type, sensor_name, attrs, samp_attrs,
			     locs, pols */);

extern void GenerateEventFormat(/* numSensors, sensors, startSensorNumber,
				   eventFormatFileName, sensorHFileName */ );
extern void GenerateSensorHFile(/* numSensors, sensors, startSensorNumber,
				   sensorHFile, sensorHFileName */);
    
extern void GenerateSensorCodeFile(/* numSensors, sensors, startSensorNumber,
				      sensorCodeFile, sensorHFileName */);

extern void GenerateSensorExtHFile(/* numSensors, sensors, startSensorNumber,
				      sensorExtHFile */);

SensorP		sensors = NULL;
int		numSensors = 0;

int             errors = 0;
%}
%union{
    SensorP        NTSensorType;
    AttributeP     NTAttributeType;
    AttributeType  NTSimpleType;
    LocationP      NTLocationType;
    PolicyP	   NTPolicyType;
    NumberListP    NTNumberListType;
    char*	   NTStringType;
    int		   NTint;
    char	   nothing;
}


%token EVENTSENSOR STATESENSOR SAMPLESENSOR SENSOR
%token SAMPATTR ATTRIBUTE LOCATION POLICY

%token DEFAULT USER_SPECIFY VERBATIM
%token BEFORE_PROC_CALL AFTER_PROC_CALL
%token BEGINNING_PROC END_PROC
%token BEFORE_SEND  AFTER_SEND
%token BEFORE_RECEIVE AFTER_RECEIVE
%token BEFORE_SH_ACCESS AFTER_SH_ACCESS
%token AT_LINES IN_FILE

%token EXPRESSION NUMBER ID

%token CHAR SHORT INT LONG
%token UCHAR USHORT UINT ULONG
%token FLOAT DOUBLE
%token STRING CHARSTAR

%token COLON COMMA SEMICOLON EQUAL
%token LPARAN RPARAN LBRACKET RBRACKET


%type <NTSensorType>		_sensorspec _sensors
%type <NTSensorType>   		_statedef _eventdef _sampledef

%type <NTAttributeType>     	_attributes _sampleattributes
%type <NTAttributeType>		_attribute_list _attribute
%type <NTint>			_attrtype
%type <NTSimpleType>		_simple_type _string_type
%type <NTint>			_array_size

%type <NTStringType>		_attrexpr _attribute_expression
%type <NTStringType>		_attribute_expr_part
%type <NTStringType>		_name

%type <NTLocationType>		_locations _location_values _location_value
%type <NTPolicyType>		_policies _policy_values _policy_value
%type <NTNumberListType>	_lines _line_number

%%


_sensorspec :
        _sensors
	{
	    SensorP  p;
	    
	    p = sensors = $1;
	    while (p) {numSensors++; p = p->next;}
	    $$ = $1;
	}
        ;


_sensors :
        _eventdef
        {
	    if (!ValidSensor($1, NULL)) errors++;
	    $$ = $1;
        }
        |
        _statedef
	{
	    if (!ValidSensor($1, NULL)) errors++;
	    $$ = $1;
	}
        |
	_sampledef
	{
	    if (!ValidSensor($1, NULL)) errors++;
	    $$ = $1;
	}
	|
        _eventdef _sensors
        {
	    if (!ValidSensor($1, $2)) errors++;
	    $2->prev = $1;
	    $1->next = $2;

	    $$ = $1;
        }
        |
        _statedef _sensors
	{
	    if (!ValidSensor($1, $2)) errors++;
	    $2->prev = $1;
	    $1->next = $2;

	    $$ = $1;
        }
        |
        _sampledef _sensors
	{
	    if (!ValidSensor($1, $2)) errors++;
	    $2->prev = $1;
	    $1->next = $2;

	    $$ = $1;
        }
        ;


_eventdef:
	EVENTSENSOR _name LPARAN _attributes _locations _policies RPARAN
	_OPTIONAL_SEMI
        {
	    $$ = SensorInit(EVENT_SENSOR, $2, $4, NULL, $5, $6);
        }
	;


_eventdef:
	SENSOR _name LPARAN _attributes  _locations  _policies RPARAN
  	_OPTIONAL_SEMI
        {
	    $$ = SensorInit(EVENT_SENSOR, $2, $4, NULL, $5, $6);
        }
	;


_statedef:
        STATESENSOR _name LPARAN _attributes  _locations  _policies RPARAN
  	_OPTIONAL_SEMI
        {
	    $$ = SensorInit(STATE_SENSOR, $2, $4, NULL, $5, $6);
        }
        ;


_sampledef:
	SAMPLESENSOR _name LPARAN _attributes  _sampleattributes
	_locations  _policies RPARAN _OPTIONAL_SEMI
        {
	    $$ = SensorInit(SAMPLE_SENSOR, $2, $4, $5, $6, $7);
        }
	;



_attributes:
	{
	    $$ = NULL;
	}
	|
	ATTRIBUTE  LPARAN  _attribute_list  RPARAN _OPTIONAL_SEMI
	{
	    $$ = $3;
	}
	;

_sampleattributes:
	SAMPATTR  LPARAN  _attribute_list  RPARAN _OPTIONAL_SEMI
	{
	    $$ = $3;
	}
	;


_attribute_list:
	_attribute
        {
	    $$ = $1;
        }
        |
        _attribute _attribute_list
        {
	    if (!ValidAttribute($1, $2)) errors++;
	    $2->prev = $1;
	    $1->next = $2;

	    $$ = $1;
        }
	;


_attribute:
	_attrtype _name _attribute_expr_part
	{
	    short      type, array;

	    type = $1 % MAX_BASIC_TYPE;
	    array = $1 / MAX_BASIC_TYPE + 1;

	    $$ = AttributeInit(type, array, $2, $3);
	} 
	;

_attribute_expr_part:
	SEMICOLON
	{
	    StringInit($$, "DEFAULT");
	}
	|
	COLON _attrexpr SEMICOLON
	{
	    $$ = $2;
	}
	|
	EQUAL _attrexpr SEMICOLON
	{
	    $$ = $2;
	}
	;

_attrtype :
	  _simple_type
	  {
              $$ = (int) $1;
          }
          |
          _simple_type LBRACKET _array_size RBRACKET
          {
	      int type;

	      type = (int) $1 + $3 * MAX_BASIC_TYPE;
	      $$ = type;
          }
	  |
	  _string_type
	  {
	      $$ = (int) $1;
	  }
          ;

_attrexpr :
	USER_SPECIFY
        {
	    StringInit($$, "USER_SPECIFY");
        }
	|
	DEFAULT
        {
	    StringInit($$, "DEFAULT");
        }
        |
	_attribute_expression
        {
	    $$ = $1;
        } 
        ;


_attribute_expression:
	EXPRESSION
        {
            StringInit($$, expression);
        }
        ;


_array_size:
	NUMBER
	{
	    $$ = number;
	}
	;


_simple_type:
	CHAR
	{
	    $$ = CHAR_TYPE;
	}
	|
	SHORT
	{
	    $$ = SHORT_TYPE;
	}
	|
	INT
	{
	    $$ = INT_TYPE;
	}
	|
	LONG
	{
	    $$ = LONG_TYPE;
	}
	|
	UCHAR
	{
	    $$ = UCHAR_TYPE;
	}
	|
	USHORT
	{
	    $$ = USHORT_TYPE;
	}
	|
	UINT
	{
	    $$ = UINT_TYPE;
	}
	|
	ULONG
	{
	    $$ = ULONG_TYPE;
	}
	|
	FLOAT
	{
	    $$ = FLOAT_TYPE;
	}
	|
	DOUBLE
	{
	    $$ = DOUBLE_TYPE;
	}
        ;


_string_type:
	STRING
	{
	    $$ = STRING_TYPE;
	}
	|
	CHARSTAR
	{
	    $$ = STRING_TYPE;
	}
	;


_locations:
	{
	    $$ = NULL;
	}
	|
	LOCATION  _OPTIONAL_COLON  _location_values
	{
	    $$ = $3;
	}
	;

_location_values:
	_location_value
	{
	    $$ = $1;
	}
	|
	_location_value   _location_values
        {
	    $2->prev = $1;
	    $1->next = $2;

	    $$ = $1;
        }
	;

_location_value:
	DEFAULT  SEMICOLON
	{
	    $$ = LocationInit(DEFAULT_LOCATION, NULL, NULL);
	}
	|
	USER_SPECIFY  SEMICOLON
	{
	    $$ = LocationInit(USER_SPECIFY_LOCATION, NULL, NULL);
	}
	|
	BEFORE_PROC_CALL _name   SEMICOLON
	{
	    $$ = LocationInit(BEFORE_PROC_CALL_LOCATION, $2, NULL);
	}
	|
	AFTER_PROC_CALL _name   SEMICOLON
	{
	    $$ = LocationInit(AFTER_PROC_CALL_LOCATION, $2, NULL);
	}
	|
	BEGINNING_PROC  _name   SEMICOLON
	{
	    $$ = LocationInit(BEGINNING_PROC_LOCATION, $2, NULL);
	}
	|
	END_PROC  _name   SEMICOLON
	{
	    $$ = LocationInit(END_PROC_LOCATION, $2, NULL);
	}
	|
	BEFORE_SEND _name   SEMICOLON
	{
	    $$ = LocationInit(BEFORE_SEND_LOCATION, $2, NULL);
	}
	|
	AFTER_SEND _name   SEMICOLON
	{
	    $$ = LocationInit(AFTER_SEND_LOCATION, $2, NULL);
	}
	|
	BEFORE_RECEIVE _name   SEMICOLON
	{
	    $$ = LocationInit(BEFORE_RECEIVE_LOCATION, $2, NULL);
	}
	|
	AFTER_RECEIVE _name   SEMICOLON
	{
	    $$ = LocationInit(AFTER_RECEIVE_LOCATION, $2, NULL);
	}
	|
	BEFORE_SH_ACCESS _name   SEMICOLON
	{
	    $$ = LocationInit(BEFORE_SH_ACCESS_LOCATION, $2, NULL);
	}
	|
	AFTER_SH_ACCESS _name   SEMICOLON
	{
	    $$ = LocationInit(AFTER_SH_ACCESS_LOCATION, $2, NULL);
	}
	|
	AT_LINES  _lines  IN_FILE  _name   SEMICOLON
	{
	    $$ = LocationInit(AT_LINES_LOCATION, $4, $2);
	}
	;


_lines:
	_line_number
	{
	    $$ = $1;
	}
	|
	_line_number  COMMA  _lines
        {
	    $3->prev = $1;
	    $1->next = $3;

	    $$ = $1;
        }
	;


_line_number:
	NUMBER
	{
	    NumberListInit($$, number);
	}
	;


_policies:
	{
	    $$ = NULL;
	}
	|
	POLICY _OPTIONAL_COLON _policy_values
	{
	    $$ = $3;
	}
	;

_policy_values:
	_policy_value
	{
	    $$ = $1;
	}
	|
	_policy_value  _policy_values
        {
	    $2->prev = $1;
	    $1->next = $2;

	    $$ = $1;
        }
	;

_policy_value:
	DEFAULT  SEMICOLON
	{
	    PolicyInit($$, DEFAULT_POLICY);
	}
	;


_name:
	ID
        {
            $$ = name;
        }
	;

_OPTIONAL_SEMI:
	|
	SEMICOLON
	;

_OPTIONAL_COLON:
	|
	COLON
	;

%%

#define ShowArgs \
  fprintf(stderr, "Usage %s -o <output file>\n", argv[0]); \
  fprintf(stderr, "         -event_format <evemt format file>\n"); \
  fprintf(stderr, "         -sensor_code <sensor code file>\n"); \
  fprintf(stderr, "         -hfile <sensors .h file>\n"); \
  fprintf(stderr, "         -prefix <string>  # use string as prefix instead of \"US\"\n"); \
  fprintf(stderr, "         <input file>\n");


extern void OutSensorSpecification();

int  startSensorNumber = 128;

char *sensorPrefixString = "US";

void
main(argc,argv)
int argc;
char *argv[]; 
{ 
    char   *inFileName = NULL;
    char   *eventFormatFileName = NULL;
    char   *sensorCodeFileName = NULL;
    char   *sensorHFileName = NULL;
    char   *sensorExtHFileName = NULL;
    char   *tempFileName;
    FILE   *sensorCodeFile, *sensorHFile, *sensorExtHFile;
    short  i, l;
    char   m4Command[512];
    int    commandP = 0;

    inFileName = eventFormatFileName = NULL;
    sprintf(m4Command, "m4 "); commandP = 3;
    
    /*** command line arguments processing ***/
    for (i = 1; i < argc; i++) {
	if (!strncmp(argv[i], "-event_format", 4)) {
	    /*** event format file name ***/
	    if (i == argc-1) {
		ShowArgs;
		exit(1);
	    } else  eventFormatFileName = argv[++i];

	} else if (!strncmp(argv[i], "-sensor_code", 4)) {
	    /*** sensor code file ***/
	    if (i == argc-1) {
		ShowArgs;
		exit(1);
	    } else  sensorCodeFileName = argv[++i];
	    
	} else if (!strncmp(argv[i], "-hfile", 4)) {
	    /*** sensors .h file ***/
	    if (i == argc-1) {
		ShowArgs;
		exit(1);
	    } else  sensorHFileName = argv[++i];
	} else if (!strcmp(argv[i], "-prefix")) {
	    /*** string to use as prefix ***/
	    if (i == argc-1) {
		ShowArgs;
		exit(1);
	    } else  sensorPrefixString = argv[++i];
	    /*** options for m4 ***/
	} else if (!strcmp(argv[i], "-s") || !strcmp(argv[i], "-e")
		   || !strcmp(argv[i], "-se")) {
	    sprintf(&m4Command[commandP], "%s ", argv[i]);
	    commandP += strlen(argv[i])+1;
	} else if (!strncmp(argv[i], "-B", 2) || !strncmp(argv[i], "-H", 2)
		   || !strncmp(argv[i], "-S", 2) || !strncmp(argv[i], "-T", 2)
		   || !strncmp(argv[i], "-D", 2) || !strncmp(argv[i], "-U",2)) {
	    sprintf(&m4Command[commandP], "%s ", argv[i]);
	    commandP += strlen(argv[i])+1;
	    
	} else {
	    /*** input file name ***/
	    if (inFileName) {
		ShowArgs;
		exit(1);
	    } else {
		l = strlen(argv[i]);
		inFileName = (char*) malloc(l+3);
		strcpy(inFileName, argv[i]);
		if (!((l >= 2) && (inFileName[l-1] == 's')
		      && (inFileName[l-2] == '.'))) {
		    strcat(inFileName, ".s");
		    l += 2;
		}
	    }
	}
    }

    /*** invoke m4 and output to file with name tempFileName ***/
    if (inFileName) {
	tempFileName = (char*) malloc(strlen(inFileName)+16);
	sprintf(tempFileName, ".temp_input.%s", inFileName);
	sprintf(&m4Command[commandP], "%s > %s", inFileName, tempFileName);
    } else {
	tempFileName = ".temp_input_file";
	sprintf(&m4Command[commandP], "> %s", tempFileName);
    }

    system(m4Command);
    
    /*** open input file ***/
    if (!freopen(tempFileName,"r",stdin)) {
	fprintf(stderr, "Macro expassion error detected\n");
	exit(1);
    }
    unlink(tempFileName);
    
    /*** open  output file ***/
    /*** open event format file ***/
    if (!eventFormatFileName) {
	if (!inFileName) eventFormatFileName = "-";
	else {
	    eventFormatFileName = (char*) malloc(l+14);
	    strcpy(eventFormatFileName, inFileName);
	    eventFormatFileName[l-1] = '\0';
	    strcat(eventFormatFileName, "eventformat.c");
	}
    }
    
    /*** open sensor code file ***/
    if (!sensorCodeFileName) {
	if (!inFileName) sensorCodeFile = stdout;
	else {
	    sensorCodeFileName = (char*) malloc(l+13);
	    strcpy(sensorCodeFileName, inFileName);
	    sensorCodeFileName[l-1] = '\0';
	    strcat(sensorCodeFileName, "sensorcode.c");
	}
    }

    if (sensorCodeFileName) {
	sensorCodeFile = fopen(sensorCodeFileName, "w+");
	if (!sensorCodeFile) {
	    fprintf(stderr, "Cannot open %s\n", sensorCodeFileName);
	    exit(1);
	}
    }
    
    /*** open sensor .h file ***/
    if (!sensorHFileName) {
	if (!inFileName) sensorHFile = stdout;
	else {
	    sensorHFileName = (char*) malloc(l+12);
	    strcpy(sensorHFileName, inFileName);
	    sensorHFileName[l-1] = '\0';
	    strcat(sensorHFileName, "sensors.h");
	}
    }

    if (sensorHFileName) {
	sensorHFile = fopen(sensorHFileName, "w+");
	if (!sensorHFile) {
	    fprintf(stderr, "Cannot open %s\n", sensorHFileName);
	    exit(1);
	}
    }

    /*** open sensor extern H file ***/
    if (!inFileName) sensorExtHFile = stdout;
    else {
	sensorExtHFileName = (char*) malloc(l+16);
	strcpy(sensorExtHFileName, inFileName);
	sensorExtHFileName[l-1] = '\0';
	strcat(sensorExtHFileName, "sensorextern.h");
	sensorExtHFile = fopen(sensorExtHFileName, "w+");
	if (!sensorExtHFile) {
	    fprintf(stderr, "Cannot open %s\n", sensorExtHFileName);
	    exit(1);
	}
    }
    
    if (yyparse()) {
	fprintf(stderr, "\nSyntax error detected.\n");
	exit(1);
    } else if (errors) {
	fprintf(stderr, "\n%d error(s) detected.\n", errors);
	exit(1);
    }

    GenerateSensorHFile(numSensors, sensors, startSensorNumber,
			sensorHFile, sensorHFileName);
    
    GenerateSensorCodeFile(numSensors, sensors, startSensorNumber,
		       sensorCodeFile, sensorHFileName);

    GenerateSensorExtHFile(numSensors, sensors, startSensorNumber,
			   sensorExtHFile);
    
    fclose(sensorCodeFile);
    fclose(sensorHFile);
    fclose(sensorExtHFile);

    exit(0);
}


int
yyerror()
{
    fprintf(stderr, "### syntax error at line: %d, near string: \"%s\" ###\n",
	    yy_line_num, yytext);
    return 0;
}
