/**********************************************************************
 **                                                                  **
 **                            edit_parms.cc                         **
 **                                                                  **
 **                                                                  **
 **  Written by:  Douglas C. MacKenzie                               **
 **                                                                  **
 **  Copyright 1995 - 1997, 1999 - 2003 Georgia Tech Research        **
 **  Corporation Atlanta, Georgia  30332-0415                        **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **********************************************************************/

/* $Id: EditParms.cc,v 1.1.1.1 2006/07/12 13:37:54 endo Exp $ */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <ctype.h>

#include <Xm/Xm.h>
#include <Xm/DialogS.h>
#include <Xm/PanedW.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/Form.h>
#include <Xm/SelectioB.h>
#include <Xm/LabelG.h>
#include <Xm/RowColumn.h>
#include <Xm/PushBG.h>
#include <Xm/Scale.h>
#include <Xm/ToggleB.h>

#include "load_cdl.h"
#include "popups.h"
#include "configuration.hpp"
#include "globals.h"
#include "PuDataType.h"
#include "ConstructPrintName.h"
#include "EditParms.h"
#include "TypeInfo.hpp"
#include "mission_expert.h"

struct toggle_info
{
   int            bit;
   unsigned long *status;
};

struct parm_def
{
   Widget          w;
   class TypeInfo *TypeRecord;
   Symbol         *ig;

   ~parm_def() {};
};

GTList<parm_def *> dialog_def;
static bool done;

/*-----------------------------------------------------------------------*/
static void
Ok_cb(Widget w, Widget shell)
{
   while( !dialog_def.isempty() )
   {
      parm_def *pd = dialog_def.get();

      pd->TypeRecord->UpdateFromDisplayWidget(pd->w, &pd->ig->name);
   }

   // Regenerate the drawing to get the changes
   config->regen_page();
   
   // Mark drawing as changed
   config->made_change();

   if (gMExp->isEnabled())
   {
       gMExp->madeChange();
   }
   
   // Kill us
   XtDestroyWidget(shell);
	 
   done = true;
}

/*-----------------------------------------------------------------------*/

static void
cancel_cb(Widget w, Widget shell)
{
   // Dump the old record
   while( !dialog_def.isempty() )
      delete dialog_def.get();

   // Kill us
   XtDestroyWidget(shell);

   done = true;
}

/*-----------------------------------------------------------------------*/

void
EditParms(Symbol *rec)
{
   if( rec == NULL )
   {
      warn_user("Internal Error: Null record to EditParms"); 
      return;
   }

   char buf[2048];
   char *n = ObjectName(rec);
   sprintf(buf,"Set %s Parameters", n);
   free(n);

   Widget dialog = XtVaCreatePopupShell("EditParms",
	xmDialogShellWidgetClass, main_window, 
	XmNtitle, buf,
        XmNx, 100,  
	XmNy, 100,
	NULL);

   Widget pane = XtVaCreateWidget("pane", xmPanedWindowWidgetClass, dialog,
	XmNsashWidth, 1,
	XmNsashHeight, 1,
	NULL);

   //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

   // Create control area
   Widget control = XtVaCreateWidget("",
        xmRowColumnWidgetClass, pane,
	XmNleftAttachment,	XmATTACH_FORM,
	XmNrightAttachment,	XmATTACH_FORM,
	XmNtopAttachment,	XmATTACH_FORM,
        NULL);

   // Make a modification line for each parameter
   Symbol *p;
   void *cur;
   if ((cur = rec->parameter_list.first(&p)) != NULL)
   {
      do 
      {
         if( p->name[0] == '%' )
         {
            Symbol *data_type = PuDataType(rec, p->name);
            if( data_type && data_type->TypeRecord )
            {
               // Remember relevent stuff so the ok callback can update the tree.
               parm_def *pd = new parm_def;
               pd->ig = p->input_generator;
               pd->TypeRecord = data_type->TypeRecord;

               pd->w = data_type->TypeRecord->CreateDisplayWidget(control,p);
               dialog_def.append(pd);
            }
         }
      } while ((cur = rec->parameter_list.next(&p, cur)) != NULL);
   }

   XtManageChild(control);

   //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

   // Create the action area
   Widget form2 = XtVaCreateWidget("", xmFormWidgetClass, pane, 
        XmNfractionBase, 5,
	NULL);

   // Create the OK button.
   Widget ok = XtVaCreateManagedWidget("Ok", xmPushButtonGadgetClass, form2,
	XmNtopAttachment,	XmATTACH_FORM,
	XmNbottomAttachment,	XmATTACH_FORM,
	XmNleftAttachment,	XmATTACH_POSITION,
	XmNleftPosition,	1,
	XmNrightAttachment,	XmATTACH_POSITION,
	XmNrightPosition,	2,
	XmNshowAsDefault,	True,
	XmNdefaultButtonShadowThickness, 1,
	NULL);
   XtAddCallback(ok, XmNactivateCallback, (XtCallbackProc)Ok_cb, dialog);

   // Create the Cancel button.
   Widget cancel = XtVaCreateManagedWidget("Cancel", 
	xmPushButtonGadgetClass, form2,
	XmNtopAttachment,	XmATTACH_FORM,
	XmNbottomAttachment,	XmATTACH_FORM,
	XmNleftAttachment,	XmATTACH_POSITION,
	XmNleftPosition,	3,
	XmNrightAttachment,	XmATTACH_POSITION,
	XmNrightPosition,	4,
	XmNshowAsDefault,	False,
	XmNdefaultButtonShadowThickness, 1,
	NULL);
   XtAddCallback(cancel, XmNactivateCallback, (XtCallbackProc)cancel_cb,dialog);
   XtManageChild(form2);

#if 0
   // Fix the height of the action area
   Dimension h;
   XtVaGetValues(cancel, 
	XmNheight, &h, 
	NULL);
   XtVaSetValues(form2, 
	XmNpaneMaximum, h, 
	XmNpaneMinimum, h, 
        XmNdefaultButton, ok,
	NULL);
#endif

   //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

   XtManageChild(pane);
   XtPopup(dialog, XtGrabExclusive);
   
   /* Wait for the answer */    
   done = false;
   while( !done )
   {
      XtAppProcessEvent(app, XtIMAll);
   }
}


///////////////////////////////////////////////////////////////////////
// $Log: EditParms.cc,v $
// Revision 1.1.1.1  2006/07/12 13:37:54  endo
// MissionLab 7.0
//
// Revision 1.1.1.1  2005/02/06 22:59:31  endo
// AO-FNC Project (based on mlab-6.0.01.06192003.tar.gz)
//
// Revision 1.20  2003/04/06 12:25:34  endo
// gcc 3.1.1
//
// Revision 1.19  2003/04/06 09:00:07  endo
// Updated for CBR Wizard Prototype II.
//
// Revision 1.18  2002/01/12 23:23:16  endo
// Mission Expert functionality added.
//
// Revision 1.17  2000/04/13 22:27:57  endo
// Checked in for Doug.
//
// Revision 1.16  1999/07/03 21:40:03  endo
// The point-and-click function moved to in
// src/load_cdl/TwoPoints.hpp.
//
// Revision 1.15  1999/06/02 23:38:32  xu
// Fixed a bug in the point-and-click function that crashes with dealing multiple parameters
// of a GoTo behavior.
//
// Revision 1.14  1999/06/02 01:53:11  xu
// Modified the function of middle-button-clicking on a "GoTo" behavior
// to bring up mlab to display an overlay, so the user can use mouse to
// point-and-click on the overlay and get the target coordinates easily.
//
// Revision 1.13  1997/02/14  16:53:35  zchen
// *** empty log message ***
//
// Revision 1.12  1996/10/04  20:58:17  doug
// changes to get to version 1.0c
//
// Revision 1.12  1996/10/02 21:45:20  doug
// working on pushup in states and transitions
//
// Revision 1.11  1996/05/08 15:16:39  doug
// fixing compile warnings
//
// Revision 1.10  1996/04/06  23:54:05  doug
// *** empty log message ***
//
// Revision 1.9  1996/03/08  00:46:56  doug
// *** empty log message ***
//
// Revision 1.8  1996/03/01  00:47:46  doug
// *** empty log message ***
//
// Revision 1.7  1996/02/29  01:48:49  doug
// *** empty log message ***
//
// Revision 1.6  1996/02/28  03:56:24  doug
// *** empty log message ***
//
// Revision 1.5  1996/02/27  05:01:12  doug
// *** empty log message ***
//
// Revision 1.4  1996/02/26  05:01:33  doug
// *** empty log message ***
//
// Revision 1.3  1996/02/25  01:13:14  doug
// *** empty log message ***
//
// Revision 1.2  1996/02/20  22:52:24  doug
// adding EditParms
//
// Revision 1.1  1996/02/20  22:20:48  doug
// Initial revision
//
///////////////////////////////////////////////////////////////////////
