/**********************************************************************
 **                                                                  **
 **                         cfgedit_cbrclient.h                      **
 **                                                                  **
 **  Written by:  Yoichiro Endo                                      **
 **                                                                  **
 **  This is a header file for cbrclient.cc.                         **
 **                                                                  **
 **  Copyright 2003 - 2005 Georgia Tech Research Corporation         **
 **  Atlanta, Georgia  30332-0415                                    **
 **  ALL RIGHTS RESERVED, See file COPYRIGHT for details.            **
 **                                                                  **
 **********************************************************************/
#ifndef CFGEDIT_CBRCLIENT_H
#define CFGEDIT_CBRCLIENT_H

/* $Id: cfgedit_cbrclient.h,v 1.1.1.1 2006/07/12 13:37:54 endo Exp $ */

#include <string>
#include <vector>
#include "cbrclient.h"
#include "mission_expert_types.h"
#include "cbrplanner_protocol.h"
#include "configuration.hpp"

using std::string;
using std::vector;

class CfgEditCBRClient : public CBRClient {

protected:
    static const string STRING_TASK_;
    static const string STRING_START_X_;
    static const string STRING_START_Y_;
    static const string STRING_ROBOT_CONSTRAINTS_;

    MExpLogfileInfo_t retrieveLogfileInfoFromServer_(void);
    CBRPlanner_MissionPlan_t retrieveMissionPlanFromServer_(void);
    CBRPlanner_CNPStatus_t retrieveCurrentCNPStatusFromServer_(void);
    vector<int> sendMissionPlanToServer_(
        CBRPlanner_MissionPlan_t missionplan,
        vector<CBRPlanner_Feature_t> features);
    vector<CBRPlanner_MissionPlan_t> receiveMissionPlansFromServer_(void);
    vector<CBRPlanner_MissionPlan_t> retrieveMissionPlansFromServer_(
        vector<CBRPlanner_Feature_t> features,
        int maxRating);
    vector<CBRPlanner_MissionPlan_t> retrieveMissionPlansFromServerPremissionCNP_(
        vector<CBRPlanner_Feature_t> features,
        int maxRating,
        vector<CBRPlanner_TaskConstraints_t> taskConstraintsList,
        vector<CBRPlanner_RobotConstraints_t> robotConstraintsList);
    vector<CBRPlanner_MissionPlan_t> retrieveMissionPlansFromServerRuntimeCNP_(
        vector<CBRPlanner_Feature_t> features,
        int maxRating,
        vector<CBRPlanner_RobotConstraints_t> robotConstraintsList);
    vector<CBRPlanner_RobotIDs_t> retrieveCurrentRobotIDListFromServer_(void);
    void sendFSANamesToServer_(config_fsa_lists_t fsaLists);
    void sendFeedbackToServer_(const unsigned char cmd, vector<int> dataIndexList);
    void readAndProcessServerData_(void);
    void sendLogfileInfoToServer_(vector<MExpLogfileInfo_t> logfileList);
    void sendOverlayInfoToServer_(string overlayFileName);
    void sendFeaturesToServer_(vector<CBRPlanner_Feature_t> features);
    void sendConstraintsToServer_(vector<CBRPlanner_Constraints_t> constraintsList);

public:
    CfgEditCBRClient(void);
    CfgEditCBRClient(const symbol_table<rc_chain> &rcTable);
    ~CfgEditCBRClient(void);
    vector<int> saveMissionPlan(
        vector<MExpFeature_t> globalFeatures_,
        MExpMissionTask_t task,
        double taskWeight,
        string solutionFileName);
    vector<MExpRetrievedMission_t> retrieveMissionPlan(
        string featureFileName,
        vector<MExpMissionTask_t> defaultTaskList,
        vector<MExpFeature_t> defaultGlobalFeatures,
        vector<MExpFeature_t> defaultLocalFeatures,
        double taskWeight,
        int maxRating,
        string *overlayFileName,
        int *cnpMode,
        vector<MExpRobotIDs_t> *robotIDList,
        MExpCNPStatus_t *cnpStatus,
        bool *reload);
    vector<MExpRetrievedMission_t> fixMissionPlan(
        string missionPlanFileName);
    vector<MExpStateInfo_t> retrieveStateInfoFromServer_(void);
    void saveFSANames(config_fsa_lists_t fsaLists);
    void savePositiveFeedback(vector<int> dataIndexList);
    void saveNegativeFeedback(vector<int> dataIndexList);
    void saveHighlightedStateInfo(MExpStateInfo_t highlightedStateInfo);
};

extern CfgEditCBRClient *gCfgEditCBRClient;

#endif
/**********************************************************************
 * $Log: cfgedit_cbrclient.h,v $
 * Revision 1.1.1.1  2006/07/12 13:37:54  endo
 * MissionLab 7.0
 *
 * Revision 1.7  2006/02/19 17:52:51  endo
 * Experiment related modifications
 *
 * Revision 1.6  2006/02/14 02:27:18  endo
 * gAutomaticExecution flag and its capability added.
 *
 * Revision 1.5  2006/01/30 02:47:27  endo
 * AO-FNC CBR-CNP Type-I check-in.
 *
 * Revision 1.4  2005/08/12 21:49:58  endo
 * More improvements for August demo.
 *
 * Revision 1.3  2005/07/31 03:39:43  endo
 * Robot ID specified by CNP can be now recognized by mlab during execution.
 *
 * Revision 1.2  2005/07/16 08:47:49  endo
 * CBR-CNP integration
 *
 * Revision 1.1  2005/02/07 22:25:27  endo
 * Mods for usability-2004
 *
 **********************************************************************/
